@@ -1,4 +1,9 @@
+#ifdef HAVE_CONFIG_H
 #include "config.h"
+#endif /* HAVE_CONFIG_H */
+
+#include <glib.h>
+#include <glib/gi18n.h>
 
 #include <unistd.h>
 #include <stdlib.h>
@@ -13,15 +18,15 @@
 #include <sys/ioctl.h>
 #ifdef __FreeBSD__
 #include <sys/uio.h>
-#endif
-#include <glib.h>
+#endif /* __FreeBSD__ */
 #include <signal.h>
-#include <libgnome/gnome-i18n.h>
+
+#ifdef USE_HAL
+#include <libhal.h>
+#endif
 
 #include "cd-recorder.h"
-#include "cdrecorder-marshal.h"
-
-#undef DEBUG_CDRECORD
+#include "cd-recorder-marshal.h"
 
 typedef enum {
 	PREPARING_TO_WRITE_CD,
@@ -46,14 +51,20 @@ struct CDRecorderPrivate {
 	GString *line;
 	GString *cdr_stderr;
 	gboolean changed_text;
-	gboolean send_return;
 	gboolean expect_cdrecord_to_die;
 	GList *actions;
+	gboolean dangerous;
 
 	char *last_error;
-	
+
 	GList *tracks;
 	int track_count;
+	gboolean debug;
+	gboolean can_rewrite;
+
+#ifdef USE_HAL
+	LibHalContext *ctx;
+#endif
 };
 
 /* Signals */
@@ -61,9 +72,7 @@ enum {
 	PROGRESS_CHANGED,
 	ACTION_CHANGED,
 	ANIMATION_CHANGED,
-	CANCEL_FUNC_CHANGED,
 	INSERT_CD_REQUEST,
-	INSERT_RWCD_REQUEST,
 	LAST_SIGNAL
 };
 
@@ -76,45 +85,53 @@ static void cd_recorder_init       (CDRecorder      *cdrecorder);
 
 static int cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 		CDDrive *recorder, GList *tracks,
-		int speed, gboolean eject, gboolean blank,
-		gboolean dummy_write, gboolean disc_at_once);
+		int speed, CDRecorderWriteFlags flags);
 static int cd_recorder_write_growisofs (CDRecorder *cdrecorder,
-		CDDrive *recorder, GList *tracks,
-		int speed, gboolean eject, gboolean blank,
-		gboolean dummy_write);
+		CDDrive *recorder, GList *tracks, int speed,
+		CDRecorderWriteFlags flags);
 
-GType
-cd_recorder_get_type (void)
+G_DEFINE_TYPE(CDRecorder, cd_recorder, G_TYPE_OBJECT)
+
+void
+cd_recorder_track_free (Track *track)
 {
-	static GType cd_recorder_type = 0;
-
-	if (!cd_recorder_type) {
-		static const GTypeInfo cd_recorder_info = {
-			sizeof (CDRecorderClass),
-			(GBaseInitFunc) NULL,
-			(GBaseFinalizeFunc) NULL,
-			(GClassInitFunc) cd_recorder_class_init,
-			(GClassFinalizeFunc) NULL,
-			NULL /* class_data */,
-			sizeof (CDRecorder),
-			0 /* n_preallocs */,
-			(GInstanceInitFunc) cd_recorder_init,
-		};
-
-		cd_recorder_type = g_type_register_static (G_TYPE_OBJECT,
-				"GtkPlaylist", &cd_recorder_info,
-				(GTypeFlags)0);
+	switch (track->type) {
+	case TRACK_TYPE_DATA:
+		g_free (track->contents.data.filename);
+		break;
+	case TRACK_TYPE_AUDIO:
+		g_free (track->contents.audio.filename);
+		g_free (track->contents.audio.cdtext);
+		break;
+	default:
+		g_warning ("Invalid track type %d", track->type);
 	}
 
-	return cd_recorder_type;
+	g_free (track);
 }
 
 static gboolean
-cd_needs_growisofs (CDDrive *recorder)
+can_burn_dvds (CDDrive *recorder)
 {
-	CDMediaType type;
+	if (!(recorder->type & CDDRIVE_TYPE_DVD_RW_RECORDER) &&
+	    !(recorder->type & CDDRIVE_TYPE_DVD_PLUS_R_RECORDER) &&
+	    !(recorder->type & CDDRIVE_TYPE_DVD_PLUS_RW_RECORDER)) {
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static gboolean
+cd_needs_growisofs (CDDrive *recorder, CDMediaType type, GList *tracks)
+{
+	GList *l;
+
+	/* If we cannot burn DVDs, then we don't need growisofs */
+	if (can_burn_dvds (recorder) == FALSE) {
+		return FALSE;
+	}
 
-	type = guess_media_type (recorder->device);
 	if (type == CD_MEDIA_TYPE_DVDR
 			|| type == CD_MEDIA_TYPE_DVDRW
 			|| type == CD_MEDIA_TYPE_DVD_PLUS_R
@@ -122,33 +139,30 @@ cd_needs_growisofs (CDDrive *recorder)
 		return TRUE;
 	}
 
+	/* If we have audio tracks, we're using cdrecord */
+	for (l = tracks; l != NULL; l = l->next) {
+		Track *track = l->data;
+		if (track->type == TRACK_TYPE_AUDIO) {
+			return FALSE;
+		}
+	}
+
 	return FALSE;
 }
 
-static void
-cancel_cdrecord (gpointer data)
+gboolean
+cd_recorder_cancel (CDRecorder *cdrecorder, gboolean skip_if_dangerous)
 {
-	CDRecorder *cdrecorder = (CDRecorder *) data;
+	if (!cdrecorder->priv->dangerous || !skip_if_dangerous) {
+		kill (cdrecorder->priv->pid, SIGINT);
 
-	kill (cdrecorder->priv->pid, SIGINT);
-
-	cdrecorder->priv->result = RESULT_CANCEL;
-	g_main_loop_quit (cdrecorder->priv->loop);
-}
-
-void
-cd_recorder_insert_cd_retry (CDRecorder *cdrecorder, gboolean is_reload)
-{
-	if (is_reload) {
-		if (cdrecorder->priv->send_return) {
-			write (cdrecorder->priv->cdr_stdin, "\n", 1);
-		} else {
-			kill (cdrecorder->priv->pid, SIGUSR1);
-		}
-	} else {
-		cdrecorder->priv->result = RESULT_RETRY;
+		cdrecorder->priv->result = RESULT_CANCEL;
 		g_main_loop_quit (cdrecorder->priv->loop);
+
+		return FALSE;
 	}
+
+	return cdrecorder->priv->dangerous;
 }
 
 static gboolean
@@ -164,11 +178,10 @@ growisofs_stdout_read (GIOChannel   *source,
 	status = g_io_channel_read_line (source,
 					 &line, NULL, NULL, NULL);
 
-#ifdef DEBUG_CDRECORD
-	if (line) {
+	if (line && cdrecorder->priv->debug) {
 		g_print ("growisofs stdout: %s", line);
 	}
-#endif
+
 	if (status == G_IO_STATUS_NORMAL) {
 		double percent;
 		int perc_1, perc_2;
@@ -194,9 +207,7 @@ growisofs_stdout_read (GIOChannel   *source,
 				       cd_recorder_table_signals[PROGRESS_CHANGED], 0,
 				       percent);
 		} else if (strstr (line, "About to execute") != NULL) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
-				       cancel_cdrecord, TRUE);
+			cdrecorder->priv->dangerous = TRUE;
 		}
 
 		g_free (line);
@@ -230,11 +241,10 @@ cdrecord_stdout_read (GIOChannel   *source,
 	status = g_io_channel_read_line (source,
 					 &line, NULL, NULL, NULL);
 	
-#ifdef DEBUG_CDRECORD
-	if (line) {
+	if (line && cdrecorder->priv->debug) {
 		g_print ("cdrecord stdout: %s", line);
 	}
-#endif
+
 	if (status == G_IO_STATUS_NORMAL) {
 		if (cdrecorder->priv->line) {
 			g_string_append (cdrecorder->priv->line, line);
@@ -258,10 +268,7 @@ cdrecord_stdout_read (GIOChannel   *source,
 				       percent);
 		} else if (g_str_has_prefix (line, "Re-load disk and hit <CR>") ||
 			   g_str_has_prefix (line, "send SIGUSR1 to continue")) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
-				       TRUE);
-			cdrecorder->priv->send_return = (*line == 'R');
+			g_assert_not_reached ();
 		} else if (g_str_has_prefix (line, "Fixating...")) {
 			g_signal_emit (G_OBJECT (cdrecorder),
 				       cd_recorder_table_signals[ACTION_CHANGED], 0,
@@ -272,9 +279,7 @@ cdrecord_stdout_read (GIOChannel   *source,
 				       1.0);
 			cdrecorder->priv->result = RESULT_FINISHED;
 		} else if (g_str_has_prefix (line, "Last chance to quit, ")) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
-				       cancel_cdrecord, TRUE);
+			cdrecorder->priv->dangerous = TRUE;
 		} else if (g_str_has_prefix (line, "Blanking PMA, TOC, pregap")) {
 			g_signal_emit (G_OBJECT (cdrecorder),
 				       cd_recorder_table_signals[ACTION_CHANGED], 0,
@@ -310,20 +315,15 @@ growisofs_stderr_read (GIOChannel   *source,
 
 	status = g_io_channel_read_line (source,
 					 &line, NULL, NULL, NULL);
-#ifdef DEBUG_CDRECORD
-	if (line) {
+	if (line && cdrecorder->priv->debug) {
 		g_print ("growisofs stderr: %s", line);
 	}
-#endif
 
 	/* TODO: Handle errors */
 	if (status == G_IO_STATUS_NORMAL && !cdrecorder->priv->expect_cdrecord_to_die) {
 		g_string_prepend (cdrecorder->priv->cdr_stderr, line);
 		if (strstr (line, "unsupported MMC profile") != NULL || (strstr (line, "already carries isofs") != NULL && strstr (line, "FATAL:") != NULL)) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
-				       FALSE);
-			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
+			g_assert_not_reached ();
 		} else if (strstr (line, "unable to open") != NULL || strstr (line, "unable to stat") != NULL) {
 			/* This fits the "open64" and "open"-like messages */
 			cdrecorder->priv->last_error = g_strdup (_("The recorder could not be accessed"));
@@ -331,6 +331,9 @@ growisofs_stderr_read (GIOChannel   *source,
 		} else if (strstr (line, "not enough space available") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("Not enough space available on the disc"));
 			cdrecorder->priv->result = RESULT_ERROR;
+		} else if (strstr (line, "end of user area encountered on this track") != NULL) {
+			cdrecorder->priv->last_error = g_strdup (_("The files selected did not fit on the CD"));
+			cdrecorder->priv->result = RESULT_ERROR;
 		} else if (strstr (line, "blocks are free") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("The files selected did not fit on the CD"));
 			cdrecorder->priv->result = RESULT_ERROR;
@@ -366,32 +369,31 @@ cdrecord_stderr_read (GIOChannel   *source,
 
 	status = g_io_channel_read_line (source,
 					 &line, NULL, NULL, NULL);
-#ifdef DEBUG_CDRECORD
-	if (line) {
+	if (line && cdrecorder->priv->debug) {
 		g_print ("cdrecord stderr: %s", line);
 	}
-#endif
+
 	/* TODO: Handle errors */
 	if (status == G_IO_STATUS_NORMAL && !cdrecorder->priv->expect_cdrecord_to_die) {
 		g_string_prepend (cdrecorder->priv->cdr_stderr, line);
 		if (strstr (line, "No disk / Wrong disk!") != NULL) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
-				       FALSE);
-			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
+			g_warning ("No disk in drive, shouldn't happen");
+			g_assert_not_reached ();
+		} else if (strstr (line, "This means that we are checking recorded media.") != NULL) {
+			cdrecorder->priv->last_error = g_strdup (_("The CD has already been recorded"));
+			cdrecorder->priv->result = RESULT_ERROR;
 		} else if (strstr (line, "Cannot blank disk, aborting.") != NULL) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_RWCD_REQUEST], 0,
-				       FALSE);
-			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
+			g_warning ("Right CD type wasn't loaded, shouldn't happen");
+			g_assert_not_reached ();
 		} else if (strstr (line, "Data may not fit on current disk") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("The files selected did not fit on the CD"));
+			/* FIXME should we error out in that case?
+			cdrecorder->priv->result = RESULT_ERROR; */
 		} else if (strstr (line, "Inappropriate audio coding") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("All audio files must be stereo, 16-bit digital audio with 44100Hz samples"));
+			cdrecorder->priv->result = RESULT_ERROR;
 		} else if (strstr (line, "cannot write medium - incompatible format") != NULL) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
-				       FALSE);
+			g_assert_not_reached ();
 			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
 		}
 
@@ -408,32 +410,134 @@ cdrecord_stderr_read (GIOChannel   *source,
 	return TRUE;
 }
 
+static gboolean
+media_type_matches (CDMediaType type, gboolean *needs_blank)
+{
+	*needs_blank = FALSE;
+
+	switch (type)
+	{
+	case CD_MEDIA_TYPE_ERROR:
+	case CD_MEDIA_TYPE_BUSY:
+		return FALSE;
+	case CD_MEDIA_TYPE_UNKNOWN:
+		return TRUE;
+	case CD_MEDIA_TYPE_CD:
+	case CD_MEDIA_TYPE_DVD:
+	case CD_MEDIA_TYPE_DVD_RAM:
+		return FALSE;
+	case CD_MEDIA_TYPE_CDR:
+		return TRUE;
+	case CD_MEDIA_TYPE_CDRW:
+		*needs_blank = TRUE;
+		return TRUE;
+	case CD_MEDIA_TYPE_DVDR:
+		return TRUE;
+	case CD_MEDIA_TYPE_DVDRW:
+		*needs_blank = TRUE;
+		return TRUE;
+	case CD_MEDIA_TYPE_DVD_PLUS_R:
+		return TRUE;
+	case CD_MEDIA_TYPE_DVD_PLUS_RW:
+		*needs_blank = TRUE;
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static CDMediaType
+cd_recorder_wait_for_insertion (CDRecorder *cdrecorder, CDDrive *drive,
+		gboolean *needs_blank)
+{
+	CDMediaType type;
+	gboolean reload;
+
+	reload = FALSE;
+	type = cd_drive_get_media_type (drive);
+
+	if (type == CD_MEDIA_TYPE_ERROR) {
+		reload = TRUE;
+	}
+
+	while (!media_type_matches (type, needs_blank)) {
+		gboolean res, busy_cd;
+
+		busy_cd = (type == CD_MEDIA_TYPE_BUSY);
+
+#ifdef USE_HAL
+		if (busy_cd != FALSE) {
+			/* TODO umount the disc if we find it,
+			 * and try it again */
+		}
+#endif
+
+		g_signal_emit (G_OBJECT (cdrecorder),
+				cd_recorder_table_signals[INSERT_CD_REQUEST],
+				0, reload, cdrecorder->priv->can_rewrite,
+				busy_cd, &res);
+
+		if (res == FALSE) {
+			return CD_MEDIA_TYPE_ERROR;
+		}
+
+		type = cd_drive_get_media_type (drive);
+		reload = FALSE;
+		if (type == CD_MEDIA_TYPE_UNKNOWN
+				|| type == CD_MEDIA_TYPE_ERROR) {
+			reload = TRUE;
+		}
+	}
+
+	return type;
+}
+
 int
 cd_recorder_write_tracks (CDRecorder *cdrecorder,
-		       CDDrive *recorder,
+		       CDDrive *drive,
 		       GList *tracks,
 		       int speed,
-		       gboolean eject,
-		       gboolean blank,
-		       gboolean dummy_write,
-		       gboolean disc_at_once)
+		       CDRecorderWriteFlags flags)
 		       /* TODO: GError **error */
 {
+	CDMediaType type;
+	gboolean needs_blank;
+
 	g_return_val_if_fail (tracks != NULL, RESULT_ERROR);
 	cdrecorder->priv->tracks = tracks;
 	cdrecorder->priv->track_count = g_list_length (tracks);
+	cdrecorder->priv->debug = (flags & CDRECORDER_DEBUG);
 
-	/* TODO: this could be changed so that cd_needs_growisofs()
-	   takes tracks, and will decide cdrecord/growisofs/error. */
-	if (cd_needs_growisofs (recorder)) {
+	/* Let's hope that DVD+RW and DVD-RW drives can also do CD-RWs */
+	cdrecorder->priv->can_rewrite =
+		(drive->type & CDDRIVE_TYPE_CDRW_RECORDER);
+
+	if (cdrecorder->priv->track_count > 99) {
+		cdrecorder->priv->last_error = g_strdup (_("You can only burn 99 tracks on one disc"));
+		cdrecorder->priv->result = RESULT_ERROR;
+		return cdrecorder->priv->result;
+	}
+
+	type = cd_recorder_wait_for_insertion (cdrecorder,
+			drive, &needs_blank);
+
+	if (type == CD_MEDIA_TYPE_ERROR) {
+		cdrecorder->priv->result = RESULT_CANCEL;
+		return cdrecorder->priv->result;
+	}
+
+	if (needs_blank != FALSE) {
+		flags |= CDRECORDER_BLANK;
+	}
+
+	if (cd_needs_growisofs (drive, type, tracks)) {
 		return cd_recorder_write_growisofs (cdrecorder,
-						    recorder, tracks, speed,
-						    eject, blank, dummy_write);
+						    drive, tracks, speed,
+						    flags);
 	} else {
 		return cd_recorder_write_cdrecord (cdrecorder,
-						   recorder, tracks, speed,
-						   eject, blank, dummy_write,
-						   disc_at_once);
+						   drive, tracks, speed,
+						   flags);
 	}
 }
 
@@ -442,9 +546,7 @@ cd_recorder_write_growisofs (CDRecorder *cdrecorder,
 				 CDDrive *recorder,
 				 GList *tracks,
 				 int speed,
-				 gboolean eject,
-				 gboolean blank,
-				 gboolean dummy_write)
+				 CDRecorderWriteFlags flags)
 {
 	GPtrArray *argv;
 	char *speed_str, *dev_str;
@@ -476,13 +578,13 @@ cd_recorder_write_growisofs (CDRecorder *cdrecorder,
 
 	/* Weird, innit? We tell growisofs we have a tty so it ignores
 	 * the fact that the DVD+ has an ISO fs already */
-	if (blank) {
+	if (flags & CDRECORDER_BLANK) {
 		g_ptr_array_add (argv, "-use-the-force-luke=tty");
 	}
 
 	g_ptr_array_add (argv, "-Z");
 
-	dev_str = g_strdup_printf ("%s=%s", recorder->device, t->data.filename);
+	dev_str = g_strdup_printf ("%s=%s", recorder->device, t->contents.data.filename);
 	g_ptr_array_add (argv, dev_str);
 
 	g_ptr_array_add (argv, NULL);
@@ -508,20 +610,17 @@ retry_growisofs:
 	g_signal_emit (G_OBJECT (cdrecorder),
 			cd_recorder_table_signals[ANIMATION_CHANGED], 0,
 			TRUE);
-	g_signal_emit (G_OBJECT (cdrecorder),
-			cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
-			cancel_cdrecord, FALSE);
+	cdrecorder->priv->dangerous = FALSE;
 
-#ifdef DEBUG_CDRECORD
+	if (cdrecorder->priv->debug)
 	{
-		int i;
+		guint i;
 		g_print ("launching command: ");
-		for (i = 0; argv->len; i++) {
+		for (i = 0; i < argv->len - 1; i++) {
 			g_print ("%s ", (char*)g_ptr_array_index (argv, i));
 		}
 		g_print ("\n");
 	}
-#endif
 
 	error = NULL;
 	if (!g_spawn_async_with_pipes  (NULL,
@@ -534,9 +633,9 @@ retry_growisofs:
 					&stdout_pipe,
 					&stderr_pipe,
 					&error)) {
-		int i;
+		guint i;
 		g_warning ("growisofs command failed: %s\n", error->message);
-		for (i = 0; argv->len; i++) {
+		for (i = 0; i < argv->len - 1; i++) {
 			g_print ("%s ", (char*)g_ptr_array_index (argv, i));
 		}
 		g_print ("\n");
@@ -564,9 +663,7 @@ retry_growisofs:
 					     cdrecorder);
 		g_io_channel_unref (channel);
 
-		g_signal_emit (G_OBJECT (cdrecorder),
-			       cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
-			       cancel_cdrecord, FALSE);
+		cdrecorder->priv->dangerous = FALSE;
 
 		g_main_loop_run (cdrecorder->priv->loop);
 		g_main_loop_unref (cdrecorder->priv->loop);
@@ -587,6 +684,14 @@ retry_growisofs:
 			cd_recorder_table_signals[ANIMATION_CHANGED], 0,
 			FALSE);
 
+	if (flags & CDRECORDER_EJECT && cdrecorder->priv->result == RESULT_FINISHED) {
+		char *cmd;
+
+		cmd = g_strdup_printf ("eject %s", recorder->device);
+		g_spawn_command_line_sync (cmd, NULL, NULL, NULL, NULL);
+		g_free (cmd);
+	}
+
 	return cdrecorder->priv->result;
 }
 
@@ -595,10 +700,7 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 		       CDDrive *recorder,
 		       GList *tracks,
 		       int speed,
-		       gboolean eject,
-		       gboolean blank,
-		       gboolean dummy_write,
-		       gboolean disc_at_once)
+		       CDRecorderWriteFlags flags)
 {
 	GPtrArray *argv;
 	char *speed_str, *dev_str;
@@ -619,18 +721,21 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 	}
 	dev_str = g_strdup_printf ("dev=%s", recorder->cdrecord_id);
 	g_ptr_array_add (argv, dev_str);
-	if (dummy_write) {
+	if (flags & CDRECORDER_DUMMY_WRITE) {
 		g_ptr_array_add (argv, "-dummy");
 	}
-	if (eject) {
+	if (flags & CDRECORDER_EJECT) {
 		g_ptr_array_add (argv, "-eject");
 	}
-	if (blank) {
+	if (flags & CDRECORDER_BLANK) {
 		g_ptr_array_add (argv, "blank=fast");
 	}
-	if (disc_at_once) {
+	if (flags & CDRECORDER_DISC_AT_ONCE) {
 		g_ptr_array_add (argv, "-dao");
 	}
+	if (flags & CDRECORDER_OVERBURN) {
+		g_ptr_array_add (argv, "-overburn");
+	}
 	g_ptr_array_add (argv, "-v");
 
 	l = tracks;
@@ -640,13 +745,13 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 		case TRACK_TYPE_DATA:
 			g_ptr_array_add (argv, "-data");
 			g_ptr_array_add (argv, "-nopad"); /* TODO? */
-			g_ptr_array_add (argv, track->data.filename);
+			g_ptr_array_add (argv, track->contents.data.filename);
 			break;
 		case TRACK_TYPE_AUDIO:
 			g_ptr_array_add (argv, "-copy");
 			g_ptr_array_add (argv, "-audio");
 			g_ptr_array_add (argv, "-pad");
-			g_ptr_array_add (argv, track->audio.filename);
+			g_ptr_array_add (argv, track->contents.audio.filename);
 			/* TODO: handle CD-TEXT somehow */
 			break;
 		default:
@@ -677,21 +782,17 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 	g_signal_emit (G_OBJECT (cdrecorder),
 		       cd_recorder_table_signals[ANIMATION_CHANGED], 0,
 		       TRUE);
-	g_signal_emit (G_OBJECT (cdrecorder),
-		       cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
-		       cancel_cdrecord, FALSE);
+	cdrecorder->priv->dangerous = FALSE;
 
-#ifdef DEBUG_CDRECORD
+	if (cdrecorder->priv->debug)
 	{
 		guint i;
 		g_print ("launching command: ");
-		for (i = 0; i < argv->len; i++) {
+		for (i = 0; i < argv->len - 1; i++) {
 			g_print ("%s ", (char*)g_ptr_array_index (argv, i));
 		}
 		g_print ("\n");
-		return -1;
 	}
-#endif
 
 	error = NULL;
 	if (!g_spawn_async_with_pipes  (NULL,
@@ -728,10 +829,8 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 					     cdrecord_stderr_read,
 					     cdrecorder);
 		g_io_channel_unref (channel);
-		
-		g_signal_emit (G_OBJECT (cdrecorder),
-			       cd_recorder_table_signals[CANCEL_FUNC_CHANGED], 0,
-			       cancel_cdrecord, FALSE);
+
+		cdrecorder->priv->dangerous = FALSE;
 
 		g_main_loop_run (cdrecorder->priv->loop);
 		g_main_loop_unref (cdrecorder->priv->loop);
@@ -784,8 +883,16 @@ cd_recorder_finalize (GObject *object)
 
 	g_return_if_fail (object != NULL);
 
-	g_string_free (cdrecorder->priv->cdr_stderr, TRUE);
-	/* FIXME cleanup */
+	if (cdrecorder->priv->cdr_stderr != NULL) {
+		g_string_free (cdrecorder->priv->cdr_stderr, TRUE);
+	}
+
+	if (cdrecorder->priv->line != NULL) {
+		g_string_free (cdrecorder->priv->line, TRUE);
+	}
+
+	g_free (cdrecorder->priv->last_error);
+	g_list_free (cdrecorder->priv->actions);
 
 	if (G_OBJECT_CLASS (parent_class)->finalize != NULL) {
 		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
@@ -848,15 +955,6 @@ cd_recorder_class_init (CDRecorderClass *klass)
 				NULL, NULL,
 				g_cclosure_marshal_VOID__BOOLEAN,
 				G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
-	cd_recorder_table_signals[CANCEL_FUNC_CHANGED] =
-		g_signal_new ("cancel-func-changed",
-				G_TYPE_FROM_CLASS (klass),
-				G_SIGNAL_RUN_LAST,
-				G_STRUCT_OFFSET (CDRecorderClass,
-					cancel_func_changed),
-				NULL, NULL,
-				cdrecorder_marshal_VOID__POINTER_BOOLEAN,
-				G_TYPE_NONE, 2, G_TYPE_POINTER, G_TYPE_BOOLEAN);
 	cd_recorder_table_signals[INSERT_CD_REQUEST] =
 		g_signal_new ("insert-cd-request",
 				G_TYPE_FROM_CLASS (klass),
@@ -864,17 +962,8 @@ cd_recorder_class_init (CDRecorderClass *klass)
 				G_STRUCT_OFFSET (CDRecorderClass,
 					insert_cd_request),
 				NULL, NULL,
-				g_cclosure_marshal_VOID__BOOLEAN,
-				G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
-	cd_recorder_table_signals[INSERT_RWCD_REQUEST] =
-		g_signal_new ("insert-rwcd-request",
-				G_TYPE_FROM_CLASS (klass),
-				G_SIGNAL_RUN_LAST,
-				G_STRUCT_OFFSET (CDRecorderClass,
-					insert_cd_rw_request),
-				NULL, NULL,
-				g_cclosure_marshal_VOID__BOOLEAN,
-				G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
-
+				cdrecorder_marshal_BOOLEAN__BOOLEAN_BOOLEAN_BOOLEAN,
+				G_TYPE_BOOLEAN, 3, G_TYPE_BOOLEAN,
+				G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
 }
 
